==> intel-oneapi-mkl-2022.1.0_223-1-aarch64.pkg.tar.xz: f_offload/dft/source/sp_complex_1d_async.f90 <==
!===============================================================================
! Copyright 2021-2022 Intel Corporation.
!
! This software and the related documents are Intel copyrighted  materials,  and
! your use of  them is  governed by the  express license  under which  they were
! provided to you (License).  Unless the License provides otherwise, you may not
! use, modify, copy, publish, distribute,  disclose or transmit this software or
! the related documents without Intel's prior written permission.
!
! This software and the related documents  are provided as  is,  with no express
! or implied  warranties,  other  than those  that are  expressly stated  in the
! License.
!===============================================================================

! Content:
! A simple example of asynchronous single-precision complex-to-complex in-place 
! 1D FFT using Intel(R) oneAPI Math Kernel Library (oneMKL) DFTI
!
!*****************************************************************************

include "mkl_dfti_omp_offload.f90"

program sp_complex_1d_async

  use MKL_DFTI_OMP_OFFLOAD, forget => DFTI_SINGLE, DFTI_SINGLE => DFTI_SINGLE_R
  use omp_lib, ONLY : omp_get_num_devices
  use, intrinsic :: ISO_C_BINDING
  ! Size of 1D transforms
  integer, parameter :: N1 = 16
  integer, parameter :: N2 = 64

  ! Arbitrary harmonic used to verify FFT
  integer, parameter :: H1 = 1
  integer, parameter :: H2 = N2/2

  ! Working precision is single precision
  integer, parameter :: WP = selected_real_kind(6,37)

  ! Execution status
  integer :: status1, status2, ignored_status

  ! The data arrays
  complex(WP), allocatable :: x1 (:)
  complex(WP), allocatable :: x2 (:)

  ! DFTI descriptor handle
  type(DFTI_DESCRIPTOR), POINTER :: hand1, hand2

  hand1 => null()
  hand2 => null()

  print *,"Example sp_complex_1d_async"
  print *,"Forward and backward single-precision complex-to-complex",        &
    &      " in-place 1D asynchronous transform"
  print *,"Configuration parameters:"
  print *,"DFTI_PRECISION      = DFTI_SINGLE"
  print *,"DFTI_FORWARD_DOMAIN = DFTI_COMPLEX"
  print *,"DFTI_DIMENSION      = 1"
  print '(" DFTI_LENGTHS        = /"I0"/ & /"I0"/ " )', N1, N2

  print *,"Create DFTI descriptor 1"
  status1 = DftiCreateDescriptor(hand1, DFTI_SINGLE, DFTI_COMPLEX, 1, N1)
  if (0 /= status1) goto 999
  
  print *,"Create DFTI descriptor 2"
  status2 = DftiCreateDescriptor(hand2, DFTI_SINGLE, DFTI_COMPLEX, 1, N2)
  if (0 /= status2) goto 999

  print *,"Commit DFTI descriptor 1"
  !$omp target variant dispatch
  status1 = DftiCommitDescriptor(hand1)
  !$omp end target variant dispatch
  if (0 /= status1) goto 999
  
  print *,"Commit DFTI descriptor 2"
  !$omp target variant dispatch
  status2 = DftiCommitDescriptor(hand2)
  !$omp end target variant dispatch
  if (0 /= status2) goto 999

  print *,"Allocate array for input data 1"
  allocate ( x1(N1), STAT = status1)
  if (0 /= status1) goto 999
  
  print *,"Allocate array for input data 2"
  allocate ( x2(N2), STAT = status2)
  if (0 /= status2) goto 999

  print *,"Initialize inputs for forward transform"
  call init(x1, N1, H1)
  call init(x2, N2, H2)

  !$omp target data map(tofrom:x1, x2)
  print *,"Compute forward transform 1"
  !$omp target variant dispatch use_device_ptr(x1) nowait
  status1 = DftiComputeForward(hand1, x1)
  !$omp end target variant dispatch
  print *,"Compute forward transform 2"
  !$omp target variant dispatch use_device_ptr(x2) nowait
  status2 = DftiComputeForward(hand2, x2)
  !$omp end target variant dispatch
  !$omp taskwait
  !$omp end target data
  if (0 /= status1) goto 999
  if (0 /= status2) goto 999

  print *,"Verify the result of FFT1"
  status1 = verify(x1, N1, H1)
  if (0 /= status1) goto 999
  print *,"Verify the result of FFT2"
  status2 = verify(x2, N2, H2)
  if (0 /= status2) goto 999

  print *,"Initialize inputs for backward transform"
  call init(x1, N1, -H1)
  call init(x2, N2, -H2)

  print *,"Compute backward transforms"
  !$omp target data map(tofrom:x1, x2)
  print *,"Compute forward transform 1"
  !$omp target variant dispatch use_device_ptr(x1) nowait
  status1 = DftiComputeBackward(hand1, x1)
  !$omp end target variant dispatch
  print *,"Compute forward transform 2"
  !$omp target variant dispatch use_device_ptr(x2) nowait
  status2 = DftiComputeBackward(hand2, x2)
  !$omp end target variant dispatch
  !$omp taskwait
  !$omp end target data
  if (0 /= status1) goto 999
  if (0 /= status2) goto 999

  print *,"Verify the results"
  status1 = verify(x1, N1, H1)
  status2 = verify(x2, N2, H2)
  if (0 /= status1) goto 999
  if (0 /= status2) goto 999

100 continue

  print *,"Release the DFTI descriptors"
  ignored_status = DftiFreeDescriptor(hand1)
  ignored_status = DftiFreeDescriptor(hand2)

  if (allocated(x1)) then
      print *,"Deallocate data array 1"
      deallocate(x1)
  endif
  
  if (allocated(x2)) then
      print *,"Deallocate data array 2"
      deallocate(x2)
  endif

  if (status1 == 0 .AND. status2 == 0) then
    print *,"TEST PASSED"
    call exit(0)
  else
    print *,"TEST FAILED"
    call exit(1)
  endif

999 print '("  Error, status1 = ",I0)', status1
  print '("  Error, status2 = ",I0)', status2
  goto 100

contains

  ! Compute mod(K*L,M) accurately
  pure real(WP) function moda(k,l,m)
    integer, intent(in) :: k,l,m
    integer*8 :: k8
    k8 = k
    moda = real(mod(k8*l,m),WP)
  end function moda

  ! Initialize array with harmonic H
  subroutine init(x, N, H)
    integer N, H
    complex(WP) :: x(:)

    integer k
    complex(WP), parameter :: I_TWOPI = (0.0_WP,6.2831853071795864769_WP)

    do k = 1, N
      x(k) = exp( I_TWOPI * moda(k-1, H, N) / cmplx(N) ) / cmplx(N)
    end do
  end subroutine init

  ! Verify that x(N) is unit peak at x(H)
  integer function verify(x, N, H)
    integer N, H
    complex(WP) :: x(:)

    integer k
    real(WP) err, errthr, maxerr
    complex(WP) res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 5.0 * log(real(N, WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    maxerr = 0.0_WP
    do k = 1, N
      if (mod(k-1-H,N)==0) then
        res_exp = 1.0_WP
      else
        res_exp = 0.0_WP
      end if
      res_got = x(k)
      err = abs(res_got - res_exp)
      maxerr = max(err,maxerr)
      if (.not.(err < errthr)) then
        print '("  x("I0"): "$)', k
        print '(" expected ("G14.7", "G14.7"),"$)', res_exp
        print '(" got ("G14.7", "G14.7"),"$)', res_got
        print '(" err "G10.3)', err
        print *," Verification FAILED"
        verify = 100
        return
      end if
    end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify = 0
  end function verify

end program sp_complex_1d_async

==> intel-oneapi-mkl-2023.2.0_49495-2-aarch64.pkg.tar.xz: f_offload/dft/source/sp_complex_1d_async.f90 <==
!===============================================================================
! Copyright 2021-2022 Intel Corporation.
!
! This software and the related documents are Intel copyrighted  materials,  and
! your use of  them is  governed by the  express license  under which  they were
! provided to you (License).  Unless the License provides otherwise, you may not
! use, modify, copy, publish, distribute,  disclose or transmit this software or
! the related documents without Intel's prior written permission.
!
! This software and the related documents  are provided as  is,  with no express
! or implied  warranties,  other  than those  that are  expressly stated  in the
! License.
!===============================================================================

! Content:
! A simple example of asynchronous single-precision complex-to-complex in-place 
! 1D FFT using Intel(R) oneAPI Math Kernel Library (oneMKL) DFTI
!
!*****************************************************************************

include "mkl_dfti_omp_offload.f90"

program sp_complex_1d_async

  use MKL_DFTI_OMP_OFFLOAD, forget => DFTI_SINGLE, DFTI_SINGLE => DFTI_SINGLE_R
  use omp_lib, ONLY : omp_get_num_devices
  use, intrinsic :: ISO_C_BINDING
  ! Size of 1D transforms
  integer, parameter :: N1 = 16
  integer, parameter :: N2 = 64

  ! Arbitrary harmonic used to verify FFT
  integer, parameter :: H1 = 1
  integer, parameter :: H2 = N2/2

  ! Working precision is single precision
  integer, parameter :: WP = selected_real_kind(6,37)

  ! Execution status
  integer :: status1, status2, ignored_status

  ! The data arrays
  complex(WP), allocatable :: x1 (:)
  complex(WP), allocatable :: x2 (:)

  ! DFTI descriptor handle
  type(DFTI_DESCRIPTOR), POINTER :: hand1, hand2

  hand1 => null()
  hand2 => null()

  print *,"Example sp_complex_1d_async"
  print *,"Forward and backward single-precision complex-to-complex",        &
    &      " in-place 1D asynchronous transform"
  print *,"Configuration parameters:"
  print *,"DFTI_PRECISION      = DFTI_SINGLE"
  print *,"DFTI_FORWARD_DOMAIN = DFTI_COMPLEX"
  print *,"DFTI_DIMENSION      = 1"
  print '(" DFTI_LENGTHS        = /"I0"/ & /"I0"/ " )', N1, N2

  print *,"Create DFTI descriptor 1"
  status1 = DftiCreateDescriptor(hand1, DFTI_SINGLE, DFTI_COMPLEX, 1, N1)
  if (0 /= status1) goto 999
  
  print *,"Create DFTI descriptor 2"
  status2 = DftiCreateDescriptor(hand2, DFTI_SINGLE, DFTI_COMPLEX, 1, N2)
  if (0 /= status2) goto 999

  print *,"Commit DFTI descriptor 1"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch
#endif
  status1 = DftiCommitDescriptor(hand1)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  if (0 /= status1) goto 999
  
  print *,"Commit DFTI descriptor 2"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch
#endif
  status2 = DftiCommitDescriptor(hand2)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  if (0 /= status2) goto 999

  print *,"Allocate array for input data 1"
  allocate ( x1(N1), STAT = status1)
  if (0 /= status1) goto 999
  
  print *,"Allocate array for input data 2"
  allocate ( x2(N2), STAT = status2)
  if (0 /= status2) goto 999

  print *,"Initialize inputs for forward transform"
  call init(x1, N1, H1)
  call init(x2, N2, H2)

  !$omp target data map(tofrom:x1, x2)
  print *,"Compute forward transform 1"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch nowait
#else
  !$omp target variant dispatch use_device_ptr(x1) nowait
#endif
  status1 = DftiComputeForward(hand1, x1)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  print *,"Compute forward transform 2"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch nowait
#else
  !$omp target variant dispatch use_device_ptr(x2) nowait
#endif
  status2 = DftiComputeForward(hand2, x2)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  !$omp taskwait
  !$omp end target data
  if (0 /= status1) goto 999
  if (0 /= status2) goto 999

  print *,"Verify the result of FFT1"
  status1 = verify(x1, N1, H1)
  if (0 /= status1) goto 999
  print *,"Verify the result of FFT2"
  status2 = verify(x2, N2, H2)
  if (0 /= status2) goto 999

  print *,"Initialize inputs for backward transform"
  call init(x1, N1, -H1)
  call init(x2, N2, -H2)

  print *,"Compute backward transforms"
  !$omp target data map(tofrom:x1, x2)
  print *,"Compute forward transform 1"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch nowait
#else
  !$omp target variant dispatch use_device_ptr(x1) nowait
#endif
  status1 = DftiComputeBackward(hand1, x1)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  print *,"Compute forward transform 2"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch nowait
#else
  !$omp target variant dispatch use_device_ptr(x2) nowait
#endif
  status2 = DftiComputeBackward(hand2, x2)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  !$omp taskwait
  !$omp end target data
  if (0 /= status1) goto 999
  if (0 /= status2) goto 999

  print *,"Verify the results"
  status1 = verify(x1, N1, H1)
  status2 = verify(x2, N2, H2)
  if (0 /= status1) goto 999
  if (0 /= status2) goto 999

100 continue

  print *,"Release the DFTI descriptors"
  ignored_status = DftiFreeDescriptor(hand1)
  ignored_status = DftiFreeDescriptor(hand2)

  if (allocated(x1)) then
      print *,"Deallocate data array 1"
      deallocate(x1)
  endif
  
  if (allocated(x2)) then
      print *,"Deallocate data array 2"
      deallocate(x2)
  endif

  if (status1 == 0 .AND. status2 == 0) then
    print *,"TEST PASSED"
    call exit(0)
  else
    print *,"TEST FAILED"
    call exit(1)
  endif

999 print '("  Error, status1 = ",I0)', status1
  print '("  Error, status2 = ",I0)', status2
  goto 100

contains

  ! Compute mod(K*L,M) accurately
  pure real(WP) function moda(k,l,m)
    integer, intent(in) :: k,l,m
    integer*8 :: k8
    k8 = k
    moda = real(mod(k8*l,m),WP)
  end function moda

  ! Initialize array with harmonic H
  subroutine init(x, N, H)
    integer N, H
    complex(WP) :: x(:)

    integer k
    complex(WP), parameter :: I_TWOPI = (0.0_WP,6.2831853071795864769_WP)

    do k = 1, N
      x(k) = exp( I_TWOPI * moda(k-1, H, N) / cmplx(N) ) / cmplx(N)
    end do
  end subroutine init

  ! Verify that x(N) is unit peak at x(H)
  integer function verify(x, N, H)
    integer N, H
    complex(WP) :: x(:)

    integer k
    real(WP) err, errthr, maxerr
    complex(WP) res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 5.0 * log(real(N, WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    maxerr = 0.0_WP
    do k = 1, N
      if (mod(k-1-H,N)==0) then
        res_exp = 1.0_WP
      else
        res_exp = 0.0_WP
      end if
      res_got = x(k)
      err = abs(res_got - res_exp)
      maxerr = max(err,maxerr)
      if (.not.(err < errthr)) then
        print '("  x("I0"): "$)', k
        print '(" expected ("G14.7", "G14.7"),"$)', res_exp
        print '(" got ("G14.7", "G14.7"),"$)', res_got
        print '(" err "G10.3)', err
        print *," Verification FAILED"
        verify = 100
        return
      end if
    end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify = 0
  end function verify

end program sp_complex_1d_async
